/*
Bug fixes:
5-18-99
Changed the feature line reading routine to accept lines of 1000 characters instead
of the previous 200.

This version is modified from line.gif; in this version the program loops
along the diagonals, thereby avoiding the redundency proportional to window 
length inherent in the previous version.

This ANSI c program does a dotplot comparison of two DNA sequences, comparing 
them by examining a stretch at time of length 'window', and recording a hit
if they are at least 'cutoff' percent identical for 'line_min' bases.

If the '-p' switch is set, the results are printed to standard output
at the indicated cutoff level of identity and for increasing identity,
until there are no more lines of identity (or 100% identity is reached).

If the '-gFILE_NAME' switch is set, the lines at the given level of homology
are printed as a gif file named FILE_NAME.  The lines printed extend over the
entire length of the windows giving the match which means that the line ends
may extend into bases that don't match but are part of a window above
the cutoff.  This can be seen most clearly with short sequences, like a mouse
and human cDNA being compared.

Command line format:
The first two options are the two DNA sequences to be compared.  They must
	come first, before the other command line switches.
'-w#' specifies the window over which identity is computed.
	Window length should be >10 bp.  I use up to 100 bp.  Bigger window
	sizes 'smooth' the resulting dotplot, but with a big window you run
	the risk of missing a small but significant feature, like a small exon.
'-s#' specifies the minimum length a stretch of identity has to have to get
	included in the output (text and gif).  This can be used as a filter
	to remove short, and presumeably insignificant matches from the output,
	allowing the longer matches to be seen more clearly.
'-c#' is the cutoff--the % identity the sequences have reach over the window
	of comparison to be considered matching.  Must be from 0 to 100.  
	The cutoff needs to be high enough to reduce the background of random
	matches.  50-70% is a good minimum.  Most conserved mouse and human 
	exons show up using a 67% cutoff.
'-gFILE_NAME' is optional.
'-p' is optional.
'-gFILE_NAME' FILE_NAME is the name of the gif file that is written.
<INPUT_FILE INPUT_FILE is the text file that line(2).gif writes as output.
	The '<' is the UNIX input redirection operator.  This argument must
	come last.

Example: line_to_gif -c72 -gDNA1.DNA2.72.gif <output.dna1.dna2

 *This library is needed to compile the code:
 *----------------
 *gd 1.2 is copyright 1994, 1995, Quest Protein Database Center, Cold Spring
 *Harbor Labs. Permission granted to copy and distribute this work provided
 *that this notice remains intact. Credit for the library must be given to the
 *Quest Protein Database Center, Cold Spring Harbor Labs, in all derived
 *works. This does not affect your ownership of the derived work itself, and
 *the intent is to assure proper credit for Quest, not to interfere with your
 *use of gd. If you have questions, ask. ("Derived works" includes all
 *programs that utilize the library. Credit must be given in user-visible
 *documentation.)
 *
 *gd 1.2 was written by Thomas Boutell and is currently distributed by
 *boutell.com, Inc.
 *
 *If you wish to release modifications to gd, please clear them first by
 *sending email to boutell@boutell.com; if this is not done, any modified
 *version of the gd library must be clearly labeled as such.
 *
 *The Quest Protein Database Center is funded under Grant P41-RR02188 by the
 *National Institutes of Health.
 *
 *Written by Thomas Boutell, 2/94 - 8/95.
 *
 *The GIF compression code is based on that found in the pbmplus utilities,
 *which in turn is based on GIFENCOD by David Rowley. 
 *-----------------
*/ 

/*
 *These are the header files needed:
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>

/*
 *These are the gif library include files needed:
 */
#include "gd.h"
#include "gdfonts.h"
#include "gdfontt.h"

/*
 *These are the constants.
 */
#define PAGEWIDTH 612
#define PAGELENGTH 792
#define P_MARGIN_X 36
#define P_MARGIN_Y 36
#define P_AXIS 36 
#define P_FEATURE_BOX_WIDTH 3
#define P_SQUISH_BOX_WIDTH 2

/*
 *These are the global variables:
 */
char *dna1_file;
char *dna2_file,*gif_file,*post_file,*limits,*limit_str;
char *dna1, *dna2;               
char fasta1[200],fasta2[200],feat_out1[10000],feat_out2[10000];
char *frame_str;
int window = 20;
float cutoff = 70;
int line_min = 20;
int match_rmax=100,match_cmax=19,limits_on=0,squish_on=0;
long int **match,**feature1,**feature2;
int next_match,redo_gif=0;
long int i,j,i_last,j_last,run_out;
float ident;
long int limit1=0,limit2=0;
long int dna1_len,limit1_end=0,limit2_end=0;
long int dna2_len;
char valid[]="acgtrynkmswbdhvACGTRYNKMBDHVxX";
char file_null[]="----";
char hh1[200],hh2[200],hh3[200],hh4[200],hh5[200],hh6[200],hh7[200],hh8[200];
typedef long int **OUT;
int reloop;
float horz=800,vert=1000;
typedef struct color3 {float a;float b;float c;} color3;

/*
 *These are all the internal functions: (the library functions aren't listed 
 *here.)
 */
int foo;
void initialize(OUT array,int x,int y);
float ident_percent(long int i2,long int j2);
void line(long int i2,long int j2,long int i_end, long int j_end);
int check_in_match(long int line_start);
void print_match(void);
char *read_dna(char *dna_file, char *dna,long int max_len, char *fasta, long int **feature,char *feat_out,char unknown);
void more_rows(void);
void more_cols();
void loop_over(long int i_start,long int i_end,long int j_start,long int j_end,float cutoff);
void gif(void);
void postscript(void);
float scaleh(float x,float each);
float scalev(float x,float each);
int sortcompare(void *,void *);
void get_header(void);
void fill_match(void);
void feat_regiffer(long int **feature);
void rect(FILE *pout, color3 c3, long int x1, long int y1, long int x2, long int y2);
void post_line(FILE *pout, long int x1, long int y1, long int x2, long int y2);
void post_text(FILE *pout, char *font_string, int size, int x, int y, char *text);
void post_textup(FILE *pout, char *font_string, int size, int x, int y, char *text);
void post_textupcentered(FILE *pout, char *font_string, int size, int x, int y,
char *text);
void post_textcentered(FILE *pout, char *font_strng, int size, int x, int y, char *text);
float post_scaleh(float x,float each);
float post_scalev(float x,float each);

/*
 *MAIN STARTS HERE.
 */
main(int argc, char *argv[])
{
int row,a=0,b,print_on=0,gif_on=0,post_on=0,frame_on=0,cutoff_flag=0;
long int i1,i2,max_len1=1000, max_len2=1000;
int feature1_rmax,feature1_cmax,feature2_rmax,feature2_cmax;
char replace_unknown;

/*
 *Allocates memory for the match 2-D array. Match holds the lines from the
 *input file.  The array is keyed on the bp of line starts in DNA seq. 1,
 *i.e. match[1][0] is the second line start bp in DNA seq. 1. match[1][1] is
 *the end of the line in DNA seq. 1. match[1][2] and [3] are the start and end
 *of the line in DNA seq. 2.  If there is a second line with the same DNA seq.
 *1 start value, the DNA seq. 1 end value would be in match[1][4]. [5] and [6]
 *would be the start and end in DNA seq. 2. And so on... 
 */
next_match=0;
match=malloc(match_rmax * sizeof(long int *));
  if(match == NULL)
    {
    puts("\nFailure to allocate memory for match!\n");
    exit(1);
    }
for(row = 0; row < match_rmax; row++)
    {
    match[row] = malloc(match_cmax * sizeof(long int));
    if(match[row] == NULL)
      {
      printf("\nFailure to allocate memory for match!\n");
      exit(1);
      }
    }

feature1_rmax=100;feature1_cmax=3;
feature1=malloc(feature1_rmax * sizeof(long int *));
  if(feature1 == NULL)
    {
    puts("\nFailure to allocate memory for feature1!\n");
    exit(1);
    }
for(row = 0; row < feature1_rmax; row++)
    {
    feature1[row] = malloc(feature1_cmax * sizeof(long int));
    if(feature1[row] == NULL)
      {
      printf("\nFailure to allocate memory for feature1!\n");
      exit(1);
      }
    }
feature2_rmax=100;feature2_cmax=3;
feature2=malloc(feature2_rmax * sizeof(long int *));
  if(feature2 == NULL)
    {
    puts("\nFailure to allocate memory for feature2!\n");
    exit(1);
    }
for(row = 0; row < feature2_rmax; row++)
    {
    feature2[row] = malloc(feature2_cmax * sizeof(long int));
    if(feature2[row] == NULL)
      {
      printf("\nFailure to allocate memory for feature2!\n");
      exit(1);
      }
    }

dna1=malloc(max_len1);
        if(dna1 == NULL)
                {
                printf("Can't get memory!\n");
                exit(1);
                }
dna2=malloc(max_len2);
        if(dna2 == NULL)
                {
                printf("Can't get memory!\n");
                exit(1);
                }

/*
 *This section reads in the command line arguments.
 *-c and -g must be used if -a is used (regiffing).
 *
 *two file names given if -a is not set (making a lineplot.)
 */
dna1_file=file_null;
dna2_file=file_null;
if (argc < 2) 
    {printf("Improper syntax, error 0, RTFM!\n");
    exit (1);
    }

    while (argc > 1) {
        /*
         * argv[1][1] is the actual option character.
         */
	if (argv[1][0] == '-')
	{
        switch (argv[1][1]) {
            /*
             * -w<number> set window size
             */
            case 'w':
                window= atoi(&argv[1][2]);
                break;
            case 'c':
               	cutoff = atoi(&argv[1][2]);
	  	cutoff_flag=1;
                break;
            case 's':
                line_min= atoi(&argv[1][2]);
                break;
            case 'a':
                redo_gif=1;
		break;
            case 'b':
                squish_on=1;
		break;
            case 'p':
                print_on=1;
		break;
            case 'g':
                gif_on=1;
		gif_file=&argv[1][2];
                break;
            case 't':
                post_on=1;
		post_file=&argv[1][2];
                break;
            case 'f':
		frame_on=1;
                frame_str=&argv[1][2];
                break;
            case 'l':
                limits_on=1;
                limits=&argv[1][2];
                break;
            default:
                (void)fprintf(stderr,"Bad argument%s\n", argv[1]);
	}
	}
	else
	    {
	    if (a == 0) 
		{dna1_file=argv[1];
		a++;
		}
	    else
		{
		if (a == 1) 
		    {dna2_file=argv[1];
		    a++;
		    }
		}
	    }
        /*
         * move the argument list up one
         * move the count down one
         */
        argv++;
        argc--;
    }


if (((dna1_file[0] == '-') || (dna2_file[0] == '-')) && (!redo_gif))
    {printf("Improper syntax, error 1, RTFM!\n");
    exit (1);
    }
if (((dna1_file[0] != '-') || (dna2_file[0] != '-')) && (redo_gif))
    {printf("Improper syntax, error 2, RTFM!\n");
    exit (1);
    }

/*
 *Parse the -fHORZ,VERT arguments.  Check to see if they are too small.
 */
if (frame_on == 1)
    {
    i=0;
    while (frame_str[i] != ',') {i++;};
    frame_str[i]='\0';
    horz= atoi(&frame_str[0]);
    vert= atoi(&frame_str[i+1]);
    if (horz < 800) 
        {printf("Horizontal dimension of gif too small: %d.  Default (800) will be used\n",horz);
        horz=800;
        }
    if (vert< 1000) 
        {printf("Vertical dimension of gif too small: %d.  Default (1000) will be used\n",vert);
        vert=1000;
        }
    }

if ((!gif_on) && (!post_on) && (redo_gif))
    {printf("No gif or postscript output file name given!\n  Use -g to specify gif file name, -t to specify postscript file name.\n");
    exit(1);
    }
if (((!cutoff) || (cutoff > 100)) && (redo_gif))
    {printf("No identity % given!  Use -c to specify identity %.\n");
    exit(1);
    }

initialize(match,match_rmax,match_cmax);   
initialize(feature1,feature1_rmax,feature1_cmax);
initialize(feature2,feature2_rmax,feature2_cmax);

/*
 *Now run either the if section or the else section.  The if section is for making a 
 *lineplot.  The else section is for redoing a lineplot (redrawing at a higher homology 
 *cutoff or a subsection of the original graph.
 */
if (!redo_gif)
{
cutoff /= 100;

replace_unknown='x';
dna1=read_dna(dna1_file,dna1,max_len1,fasta1,feature1,feat_out1,replace_unknown);
replace_unknown='n';
dna2=read_dna(dna2_file,dna2,max_len2,fasta2,feature2,feat_out2,replace_unknown);
dna1_len=strlen(dna1);
dna2_len=strlen(dna2);

/*
printf("%d,%d\n",dna1_len,dna2_len);
printf("%d,%d\n%s\n\n%s\nprint done.\n",dna1_len,dna2_len,dna1,dna2);
 */

reloop=0;
loop_over(0,dna1_len,0,dna2_len,cutoff);

qsort((void *) match,(size_t)next_match,(size_t)sizeof(long int *),sortcompare);
if (print_on) print_match();
if (gif_on) gif();
if (post_on) postscript();

if (((match[0][0] != 0) || (cutoff == 1)) && (print_on))
    {reloop=1;cutoff+=.05;
    if (cutoff > 1) cutoff=1;
    while (cutoff <= 1)
	{a=0;run_out=0;
            printf("\nRegions of %.0f%% or greater homology:\n",(cutoff*100));
	    printf("---------------------------------------------\n");
            for (i1=match[a][0];i1 != 0;i1=match[a][0])
                {b=1;
                for (i2=match[a][b];i2 != 0;i2=match[a][b])
		    {i_last=0;j_last=0;
		    loop_over((i1-1),i2,(match[a][b+1]-1),match[a][b+2],cutoff);
		    b+=3;
            	    }
                a++;
                }
        if (run_out == 0)
	    {printf("No matches.\n");
	    exit (0);
	    }
	if (cutoff == 1) break;
	cutoff+=.05;
	if (cutoff > 1) cutoff=1;
	}
    }
}
else
    {
/*
 *Begin section that redraws the gif from a previous lineplot's output file.
 */
get_header();
if (limits_on)
    {
    limit_str=&limits[0];
    for (a=0;limits[a] != '\0';a++)
        {if ((limits[a] == ',') || (limits[a] == ' '))
	    {
	    limits[a]='\0';
	    if (!limit1) 
		{limit1=atol(limit_str);
	        limit_str=&limits[a+1];
		continue;
		}
	    if (!limit1_end) 
		{if ((strcmp(limit_str,"end") == 0) || (strcmp(limit_str,"End") == 0) || (strcmp(limit_str,"END") == 0))
		    limit1_end=dna1_len;
		else 
		   limit1_end=atol(limit_str);
	        limit_str=&limits[a+1];
		continue;
		}
	    if (!limit2) 
		{limit2=atol(limit_str);
	        limit_str=&limits[a+1];
		continue;
		}
	    }
 	}
    if ((strcmp(limit_str,"end") == 0) || (strcmp(limit_str,"End") == 0) || (strcmp(limit_str,"END") == 0))
        limit2_end=dna2_len;
    else limit2_end=atol(limit_str);
    
    dna1_len=limit1_end;dna2_len=limit2_end;
    }
if ((limit1 >= dna1_len) || (limit2 > dna2_len))
    {printf("Limits specified wrong!  Check -l arguments!\n");
    exit(1);
    }

fill_match();
if (gif_on) gif();
if (post_on) postscript();
    }
 }
/*
 *End of main loop.
 *
 *Subroutines follow.
 */

/*
 *Parse feature list for regiffing.  Put pairs of numbers into array feature.
 *This input section isn't rigorous.  Errors or extra spaces in the
 *input could crash it.
 */

void feat_regiffer(long int **feature)
{int a,d,e=0,f,g;
char feat_line[1000],c;

while (1)
{
gets(feat_line);
if (feat_line[0] != ';') break;
    a=1;d=1;f=0;g=0;
    while (feat_line[a] != '\0')
       {if (!(isascii(feat_line[a]) && isdigit(feat_line[a])))
            {if (f=1)
                {f=0;
                feat_line[a]='\0';
                feature[e][g]=atol(&feat_line[d]);
                d=a+1;
                g++;
                if (g == 3) break;
                }
            }
        else
            {if (isascii(feat_line[a]) && isdigit(feat_line[a]))
                {if (f=0) d=a;
                f=1;
                }
            }
        a++;
        }
    e++;
}
}

/*
 *This subroutine is called by the regiffing part of the prog. to read in the 
 *header of the text input file.
 */
void get_header(void)
{int percent=0;
char hh0[200];

gets(hh1);
sscanf(hh1,"%*s%*s%*s%*s%*s%d",&dna1_len);
gets(hh2);
feat_regiffer(feature1);
gets(hh4);
sscanf(hh4,"%*s%*s%*s%*s%*s%d",&dna2_len);
gets(hh5);
feat_regiffer(feature2);
gets(hh7);
gets(hh8);
sscanf(hh8,"%d",&percent);
if (cutoff == percent)
    {
    gets(hh0);
    gets(hh0);
    gets(hh0);
    gets(hh0);
    }
while (!(cutoff == percent))
    {gets(hh0);
    sscanf(hh0,"Regions of %d",&percent);
    }
gets(hh0);
}

void fill_match(void)
{
long int i_start,i_end,j_start,j_end,trash;
int a=1,b,c;
while (a != EOF)
    {i_start=0;
    a=scanf("%d%d%d%d%d\n",&i_start,&i_end,&j_start,&j_end,&trash);
    if (i_start == 0) break;
    if (limits_on)
	{if ((i_start < limit1) && (i_end < limit1)) continue;
	if ((i_start > dna1_len) && (i_end > dna1_len)) continue;
	if ((j_start < limit2) && (j_end < limit2)) continue;
	if ((j_start > dna2_len) && (j_end > dna2_len)) continue;
	if (!((i_start>limit1)&&(i_end<dna1_len)&&(j_start>limit2)&&(j_end<dna2_len)))
	    {
	    if (i_start<limit1)
		{j_start+=(limit1-i_start);
		i_start=limit1;
		}
	    if (j_start<limit2)
		{i_start+=(limit2-j_start);
		j_start=limit2;
		}
	    if (i_end>dna1_len)
		{j_end-=(i_end-dna1_len);
		i_end=dna1_len;
		}
	    if (j_end>dna2_len)
		{i_end-=(j_end-dna2_len);
		j_end=dna2_len;
		}
	    if ((i_end-i_start) < 0) continue;
	    }
	}
                    b=check_in_match(i_start);
                    c=0;
                    if (b != 0)
                        {b--;
                	while (match[b][c] != 0)
                	    {c++;}
                	c--;
                	if ((c+4) >= match_cmax) more_cols();
                	}
                	else 
                	{b=next_match;next_match++;
                	if (next_match == match_rmax) more_rows();
                	match[b][c]=i_start;
                	}
                    match[b][c+1]=i_end;
                    match[b][c+2]=j_start;
                    match[b][c+3]=j_end;
    }

}

void loop_over(long int i_start,long int i_end,long int j_start,long int j_end,float cutoff)
{
if (reloop)
    {line((i_start+window),(j_start+window),i_end,j_end);
    }
else
    { 
    for (i=i_end;i >= (i_start + window);i--)
        {line(i,window,i_end,j_end);
        }
    for (j=(j_start+window+1);j <= j_end;j++)
        {line(window,j,i_end,j_end);
        }
    }
}


int sortcompare(void *n1,void *n2)
{long int *c,*d;
int a;
c=*(long int *)n1;
d=*(long int *)n2;
if ((c[0]-d[0]) >= 0) a=1;
else a=-1;
return a;
}


void gif(void)
{int a,b,c,d;
long int i1,i2,i_range,j_range;
float each1,each2;
float i,h1,h2,v1,v2;
int j,vert_end,color[27];
char number[40],description[100];        
struct tm *curtime;
time_t bintime;

	/* Declare the image */
        gdImagePtr im;
        /* Declare an output file */
        FILE *out;

if (redo_gif)
    {i_range=(dna1_len-limit1+1);
    j_range=(dna2_len-limit2+1);
    strcpy(fasta1,hh2);
    strcpy(fasta2,hh5);
    }
else 
    {i_range=dna1_len;
    j_range=dna2_len;
    };

if (match[0][0] == 0) return;

        /* Allocate the image: horz pixels across by vert pixels tall */
        im = gdImageCreate(horz, vert);

        /* Allocate the color white (red, green and blue all minimum).
                Since this is the first color in a new image, it will
                be the background color. */
        color[0]= gdImageColorAllocate(im, 255, 255,255);

        /* Allocate the color black (red, green and blue all maximum). */
        color[1]= gdImageColorAllocate(im, 0,0,0);
        color[2]= gdImageColorAllocate(im, 255,0,0);
        color[3]= gdImageColorAllocate(im, 255,255,0);
        color[4]= gdImageColorAllocate(im, 0,255,0);
        color[5]= gdImageColorAllocate(im, 0,255,255);
        color[6]= gdImageColorAllocate(im, 0,0,255);
        color[7]= gdImageColorAllocate(im, 255,0,255);
/*
 *Allocate colors 8-26, the non-primary colors.
 */
a=-127;b=0;c=0;d=8;
while (d != 27)
    {a+=127;
    if (a > 254)
	{a=0;
	b+=127;
	if (b > 254)
	    {b=0;
	    c+=127;
	    }
	}
    if (!((a != 127) && (b != 127) && (c != 127)))
	{color[d]= gdImageColorAllocate(im,a,b,c);
	d++;
	}
    }
	
if (i_range<=1000000) each1=100000;
if (i_range<=900000) each1=90000;
if (i_range<=800000) each1=80000;
if (i_range<=700000) each1=70000;
if (i_range<=600000) each1=60000;
if (i_range<=500000) each1=50000;
if (i_range<=400000) each1=40000;
if (i_range<=300000) each1=30000;
if (i_range<=200000) each1=20000;
if (i_range<=190000) each1=19000;
if (i_range<=180000) each1=18000;
if (i_range<=170000) each1=17000;
if (i_range<=160000) each1=16000;
if (i_range<=150000) each1=15000;
if (i_range<=140000) each1=14000;
if (i_range<=130000) each1=13000;
if (i_range<=120000) each1=12000;
if (i_range<=110000) each1=11000;
if (i_range<=100000) each1=10000;
if (i_range<=90000) each1=9000;
if (i_range<=80000) each1=8000;
if (i_range<=70000) each1=7000;
if (i_range<=60000) each1=6000;
if (i_range<=50000) each1=5000;
if (i_range<=40000) each1=4000;
if (i_range<=30000) each1=3000;
if (i_range<=20000) each1=2000;
if (i_range<=10000) each1=1000;
if (i_range<=9000) each1=900;
if (i_range<=8000) each1=800;
if (i_range<=7000) each1=700;
if (i_range<=6000) each1=600;
if (i_range<=5000) each1=500;
if (i_range<=4000) each1=400;
if (i_range<=3000) each1=300;
if (i_range<=2000) each1=200;
if (i_range<=1000) each1=100;
if (i_range<=500) each1=50;
if (j_range<=1000000) each2=100000;
if (j_range<=900000) each2=90000;
if (j_range<=800000) each2=80000;
if (j_range<=700000) each2=70000;
if (j_range<=600000) each2=60000;
if (j_range<=500000) each2=50000;
if (j_range<=400000) each2=40000;
if (j_range<=300000) each2=30000;
if (j_range<=200000) each2=20000;
if (j_range<=190000) each2=19000;
if (j_range<=180000) each2=18000;
if (j_range<=170000) each2=17000;
if (j_range<=160000) each2=16000;
if (j_range<=150000) each2=15000;
if (j_range<=140000) each2=14000;
if (j_range<=130000) each2=13000;
if (j_range<=120000) each2=12000;
if (j_range<=110000) each2=11000;
if (j_range<=100000) each2=10000;
if (j_range<=90000) each2=9000;
if (j_range<=80000) each2=8000;
if (j_range<=70000) each2=7000;
if (j_range<=60000) each2=6000;
if (j_range<=50000) each2=5000;
if (j_range<=40000) each2=4000;
if (j_range<=30000) each2=3000;
if (j_range<=20000) each2=2000;
if (j_range<=10000) each2=1000;
if (j_range<=9000) each2=900;
if (j_range<=8000) each2=800;
if (j_range<=7000) each2=700;
if (j_range<=6000) each2=600;
if (j_range<=5000) each2=500;
if (j_range<=4000) each2=400;
if (j_range<=3000) each2=300;
if (j_range<=2000) each2=200;
if (j_range<=1000) each2=100;
if (j_range<=500) each2=50;

/* 
 *Print out the feature boxes. First vertical, for DNA seq. 1,
 *then horz., for DNA seq. 2.
 */
a=0;
while (feature1[a][0] != 0)
    {
    if ((limits_on) && ((feature1[a][0] > limit1_end) || (feature1[a][1] < limit1)))
        {goto skip3;
        };
    if (limits_on)
        {
    if ((feature1[a][0] <= limit1) && (feature1[a][1] >= limit1_end))
        {feature1[a][0]=limit1;
        feature1[a][1]=limit1_end;
        }
    else if ((feature1[a][0] > limit1) && (feature1[a][1] > limit1_end))
        {
        feature1[a][1]=limit1_end;
        }
    else if ((feature1[a][0] < limit1) && (feature1[a][1] < limit1_end))
        {
        feature1[a][0]=limit1;
        };
	};
    v1=scalev(feature1[a][0],each1);v2=scalev(feature1[a][1],each1);
    gdImageFilledRectangle(im,45,v1,49,v2,color[feature1[a][2]]);
    skip3:
    a++;
    };

a=0;
while (feature2[a][0] != 0)
    {
    if ((limits_on) && ((feature2[a][0] > limit2_end) || (feature2[a][1] < limit2)))
        {goto skip4;
        };
    if (limits_on)
        {
    if ((feature2[a][0] <= limit2) && (feature2[a][1] >= limit2_end))
        {feature2[a][0]=limit2;
        feature2[a][1]=limit2_end;
        }
    else if ((feature2[a][0] > limit2) && (feature2[a][1] > limit2_end))
        {
        feature2[a][1]=limit2_end;
        }
    else if ((feature2[a][0] < limit2) && (feature2[a][1] < limit2_end))
        {
        feature2[a][0]=limit2;
        };
	};
    h1=scaleh(feature2[a][0],each2);h2=scaleh(feature2[a][1],each2);
    gdImageFilledRectangle(im,h1,45,h2,49,color[feature2[a][2]]);
    skip4:
    a++;
    };

for (i=1;i<=10;i++)
    {
    gdImageLine(im, 43, (((i*((vert-175)/10)))+50), 50, (((i*((vert-175)/10)))+50), color[1]);
    gdImageLine(im,50,(((i-1)*((vert-175)/10))+50), 50, (((i*((vert-175)/10)))+50),color[1]);
    gcvt((each1*i)+limit1,8,number);
    gdImageStringUp(im,gdFontTiny,35,(((i*((vert-175)/10)))+50)+(strlen(number)*gdFontTiny->w/2),number,color[1]);
    if ((each1*(i+1)) >= i_range) break;
    }
    gdImageLine(im, 43, (((i_range/each1)*((vert-175)/10))+50), 50, (((i_range/each1)*((vert-175)/10))+50), color[1]);
    gdImageLine(im,50,((i*((vert-175)/10))+50), 50, (((i_range/each1)*((vert-175)/10))+50),color[1]);

if (limits_on)
   {gcvt((float)limit1,8,number);
   gdImageStringUp(im,gdFontTiny,35,(40+(strlen(number)*gdFontTiny->w)),number,color[1]);
   }

gdImageStringUp(im,gdFontSmall,5,((((vert-175)/10*i)+(strlen(fasta1)*gdFontSmall->w))/2)+50,fasta1,color[1]);

vert_end=(((i+1)*((vert-175)/10))+50+95);

if (redo_gif)
    {gdImageString(im,gdFontSmall,60,(vert_end-90),hh1,color[1]);
gdImageString(im,gdFontSmall,70,(vert_end-79),hh2,color[1]);
gdImageString(im,gdFontSmall,60,(vert_end-68),hh4,color[1]);
gdImageString(im,gdFontSmall,70,(vert_end-57),hh5,color[1]);
gdImageString(im,gdFontSmall,60,(vert_end-46),hh7,color[1]);
sprintf(description,"%.0f%% ",cutoff);
strcat(description,(hh8 + (sizeof(char)*3)));
gdImageString(im,gdFontSmall,65,(vert_end-35),description,color[1]);
if (limits_on)
    {sprintf(description,"Only the portion of the graph from %d bp to %d bp of DNA seq. 1\n",limit1,dna1_len);
   gdImageString(im,gdFontSmall,60,(vert_end-24),description,color[1]);
    sprintf(description,"and from %d bp to %d bp of DNA seq. 2 are graphed.\n",limit2,dna2_len);
   gdImageString(im,gdFontSmall,65,(vert_end-13),description,color[1]);
    }
    }
else
    {
sprintf(description,"The y-axis DNA seq. is %d bp, in file %s:",dna1_len,dna1_file);
gdImageString(im,gdFontSmall,70,(vert_end-90),description,color[1]);
gdImageString(im,gdFontSmall,80,(vert_end-65),fasta1,color[1]);
sprintf(description,"The x-axis DNA seq. is %d bp, in file %s:",dna2_len,dna2_file);
gdImageString(im,gdFontSmall,70,(vert_end-50),description,color[1]);
gdImageString(im,gdFontSmall,80,(vert_end-35),fasta2,color[1]);
sprintf(description,"Comparing the DNA seqs. using a %d bp window, returning regions of",window);
gdImageString(im,gdFontSmall,70,(vert_end-20),description,color[1]);
sprintf(description,"%.0f%% or greater homology sustained over at least %d bp.\n\n",(cutoff*100),line_min);
gdImageString(im,gdFontSmall,80,(vert_end-10),description,color[1]);
    }

time(&bintime);
curtime=localtime(&bintime);
sprintf(description,"%s\n",asctime(curtime));
gdImageString(im,gdFontSmall,0,0,description,color[1]);

for (i=1;i<=10;i++)
    {
    gdImageLine(im,((i*((horz-90)/10))+50), 43, ((i*((horz-90)/10))+50),50, color[1]);
    gdImageLine(im,(((i-1)*((horz-90)/10))+50), 50, ((i*((horz-90)/10))+50),50,color[1]);
    gcvt((each2*i+limit2),8,number);
    gdImageString(im,gdFontTiny,((i*((horz-90)/10))+50)-(strlen(number)*gdFontTiny->w/2),35,number,color[1]);
    if ((each2*(i+1)) >= j_range) break;
    }
    gdImageLine(im,(((j_range/each2)*((horz-90)/10))+50), 43, (((j_range/each2)*((horz-90)/10))+50),50, color[1]);
    gdImageLine(im,((i*((horz-90)/10))+50), 50, (((j_range/each2)*((horz-90)/10))+50),50,color[1]);

if (limits_on)
   {gcvt((float)limit2,8,number);
   gdImageString(im,gdFontTiny,40,35,number,color[0]);
   }

gdImageString(im,gdFontSmall,((((horz-90)/10*i)-(strlen(fasta2)*gdFontSmall->w))/2)+50,8,fasta2,color[1]);

for (i=1;i<=100;i++)
    {
    if (((each1/10)*i) >= i_range) break;
    gdImageLine(im, 48, ((i*((vert-175)/100))+50), 50, ((i*((vert-175)/100))+50),color[1]);
    }
for (i=1;i<=100;i++)
    {
    if (((each2/10)*i) >= j_range) break;
    gdImageLine(im,((i*((horz-90)/100))+50), 48, ((i*((horz-90)/100))+50),50, color[1]);
    }

/*
 *The part that prints out the match lines.
 *
 */

a=0;
for (i1=match[a][0];i1 != 0;i1=match[a][0])
    {b=1;
    for (i2=match[a][b];i2 != 0;i2=match[a][b])
	{h1=scaleh(match[a][b+1],each2);h2=scaleh(match[a][b+2],each2);
	v1=scalev(i1,each1);v2=scalev(i2,each1);
	if (squish_on)
	    {gdImageFilledRectangle(im,h1,51,h2,53,color[1]);
	    gdImageFilledRectangle(im,51,v1,53,v2,color[1]);
	    }
	gdImageLine(im,h1,v1,h2,v2,color[1]);
        b+=3;
        }
    a++;
    }

        /* Open a file for writing. "wb" means "write binary", important
                under MSDOS, harmless under Unix. */
        out = fopen(gif_file, "wb");

        /* Output the image to the disk file. */
        gdImageGif(im, out);

        /* Close the file. */
        fclose(out);

        /* Destroy the image in memory. */
        gdImageDestroy(im);

}

void postscript(void)
{
int a,b,c;
long int i1,i2,i_range,j_range;
float each1,each2;
float i,h1,h2,v1,v2;
int d,j,vert_end;
char font_1st[40];
char number[40],description[100];
struct tm *curtime;
time_t bintime;
color3 color[27];

        /* Declare an output file */
FILE *pout;
        
	/* Open a file for writing. "wb" means "write binary", important
                under MSDOS, harmless under Unix. */
if((pout = fopen(post_file, "w")) == NULL)
                {
                printf("Can't open postscript file for writing!\n");
                /* exit or return */
                exit(1);
		}

sprintf(font_1st,"Times-Roman");
horz=horz/800*(PAGEWIDTH-(2*P_MARGIN_X));vert=vert/1000*(PAGELENGTH-(2*P_MARGIN_Y));

if (redo_gif)
    {i_range=(dna1_len-limit1+1);
    j_range=(dna2_len-limit2+1);
    strcpy(fasta1,hh2);
    strcpy(fasta2,hh5);
    }
else 
    {i_range=dna1_len;
    j_range=dna2_len;
    };

/*
 *Print postscript header
 */
fprintf(pout,"%%!\n");
fprintf(pout,"%%Output from lineplot.  An averaged dotplot of 2 DNA sequences.\n");
fprintf(pout,"%%(c) Jim Lund 1997.\n");
fprintf(pout,"%%\n");

/*
 *Define postscript functions and variables.
 *
 */
fprintf(pout,"[1 0 0 -1 0 792] concat\n");
fprintf(pout,"%d %d translate\n",P_MARGIN_X,P_MARGIN_Y);
fprintf(pout,".1 setlinewidth\n");
fprintf(pout,"/x1 0 def /y1 0 def /x2 0 def /y2 0 def\n");
fprintf(pout,"/Rec {newpath /y2 exch def /x2 exch def /y1 exch def /x1 exch def x1 y1 moveto x2 y1 lineto x2 y2 lineto x1 y2 lineto closepath fill } bind def\n");

 
if (match[0][0] == 0) return;

        color[0].a=1;color[0].b=1;color[0].c=1;
        color[1].a=0;color[1].b=0;color[1].c=0;
        color[2].a=1;color[2].b=0;color[2].c=0;
        color[3].a=1;color[3].b=1;color[3].c=0;
        color[4].a=0;color[4].b=1;color[4].c=0;
        color[5].a=0;color[5].b=1;color[5].c=1;
        color[6].a=0;color[6].b=0;color[6].c=1;
        color[7].a=1;color[7].b=0;color[7].c=1;
/*
 *Allocate colors 8-26, the non-primary colors.
 */
a=-127;b=0;c=0;d=8;
while (d != 27)
    {a+=127;
    if (a > 254)
	{a=0;
	b+=127;
	if (b > 254)
	    {b=0;
	    c+=127;
	    }
	}
    if (!((a != 127) && (b != 127) && (c != 127)))
	{color[d].a=((float)a/256);color[d].b=((float)b/256);color[d].c=((float)c/256);
	d++;
	}
    }
	
	
if (i_range<=1000000) each1=100000;
if (i_range<=900000) each1=90000;
if (i_range<=800000) each1=80000;
if (i_range<=700000) each1=70000;
if (i_range<=600000) each1=60000;
if (i_range<=500000) each1=50000;
if (i_range<=400000) each1=40000;
if (i_range<=300000) each1=30000;
if (i_range<=200000) each1=20000;
if (i_range<=190000) each1=19000;
if (i_range<=180000) each1=18000;
if (i_range<=170000) each1=17000;
if (i_range<=160000) each1=16000;
if (i_range<=150000) each1=15000;
if (i_range<=140000) each1=14000;
if (i_range<=130000) each1=13000;
if (i_range<=120000) each1=12000;
if (i_range<=110000) each1=11000;
if (i_range<=100000) each1=10000;
if (i_range<=90000) each1=9000;
if (i_range<=80000) each1=8000;
if (i_range<=70000) each1=7000;
if (i_range<=60000) each1=6000;
if (i_range<=50000) each1=5000;
if (i_range<=40000) each1=4000;
if (i_range<=30000) each1=3000;
if (i_range<=20000) each1=2000;
if (i_range<=10000) each1=1000;
if (i_range<=9000) each1=900;
if (i_range<=8000) each1=800;
if (i_range<=7000) each1=700;
if (i_range<=6000) each1=600;
if (i_range<=5000) each1=500;
if (i_range<=4000) each1=400;
if (i_range<=3000) each1=300;
if (i_range<=2000) each1=200;
if (i_range<=1000) each1=100;
if (i_range<=500) each1=50;
if (j_range<=1000000) each2=100000;
if (j_range<=900000) each2=90000;
if (j_range<=800000) each2=80000;
if (j_range<=700000) each2=70000;
if (j_range<=600000) each2=60000;
if (j_range<=500000) each2=50000;
if (j_range<=400000) each2=40000;
if (j_range<=300000) each2=30000;
if (j_range<=200000) each2=20000;
if (j_range<=190000) each2=19000;
if (j_range<=180000) each2=18000;
if (j_range<=170000) each2=17000;
if (j_range<=160000) each2=16000;
if (j_range<=150000) each2=15000;
if (j_range<=140000) each2=14000;
if (j_range<=130000) each2=13000;
if (j_range<=120000) each2=12000;
if (j_range<=110000) each2=11000;
if (j_range<=100000) each2=10000;
if (j_range<=90000) each2=9000;
if (j_range<=80000) each2=8000;
if (j_range<=70000) each2=7000;
if (j_range<=60000) each2=6000;
if (j_range<=50000) each2=5000;
if (j_range<=40000) each2=4000;
if (j_range<=30000) each2=3000;
if (j_range<=20000) each2=2000;
if (j_range<=10000) each2=1000;
if (j_range<=9000) each2=900;
if (j_range<=8000) each2=800;
if (j_range<=7000) each2=700;
if (j_range<=6000) each2=600;
if (j_range<=5000) each2=500;
if (j_range<=4000) each2=400;
if (j_range<=3000) each2=300;
if (j_range<=2000) each2=200;
if (j_range<=1000) each2=100;
if (j_range<=500) each2=50;


/* 
 *Print out the feature boxes. First vertical, for DNA seq. 1,
 *then horz., for DNA seq. 2.
 */
a=0;
while (feature1[a][0] != 0)
    {
    if ((limits_on) && ((feature1[a][0] > limit1_end) || (feature1[a][1] < limit1)))
	{goto skip1;
	};
    if (limits_on)
	{
    if ((feature1[a][0] <= limit1) && (feature1[a][1] >= limit1_end))
	{feature1[a][0]=limit1;
	feature1[a][1]=limit1_end;
	}
    else if ((feature1[a][0] > limit1) && (feature1[a][1] > limit1_end))
	{
	feature1[a][1]=limit1_end;
	}
    else if ((feature1[a][0] < limit1) && (feature1[a][1] < limit1_end))
	{
	feature1[a][0]=limit1;
	};
	};
    v1=post_scalev(feature1[a][0],each1);v2=post_scalev(feature1[a][1],each1);
    rect(pout,color[feature1[a][2]],(P_AXIS-1-P_FEATURE_BOX_WIDTH),v1,(P_AXIS-1),v2);
    skip1:
    a++;
    };

a=0;
while (feature2[a][0] != 0)
    {
    if ((limits_on) && ((feature2[a][0] > limit2_end) || (feature2[a][1] < limit2)))
        {goto skip2;
        };
    if (limits_on)
        {
    if ((feature2[a][0] <= limit2) && (feature2[a][1] >= limit2_end))
        {feature2[a][0]=limit2;
        feature2[a][1]=limit2_end;
        }
    else if ((feature2[a][0] > limit2) && (feature2[a][1] > limit2_end))
        {
        feature2[a][1]=limit2_end;
        }
    else if ((feature2[a][0] < limit2) && (feature2[a][1] < limit2_end))
        {
        feature2[a][0]=limit2;
        };
	};
    h1=post_scaleh(feature2[a][0],each2);h2=post_scaleh(feature2[a][1],each2);
    rect(pout,color[feature2[a][2]],h1,(P_AXIS-1-P_FEATURE_BOX_WIDTH),h2,(P_AXIS-1));
    skip2:
    a++;
    };

for (i=1;i<=10;i++)
    {
    post_line(pout, (P_AXIS-7), (((i*((vert-(P_AXIS+125))/10)))+P_AXIS), P_AXIS, (((i*((vert-(P_AXIS+125))/10)))+P_AXIS));
    post_line(pout,P_AXIS,(((i-1)*((vert-(P_AXIS+125))/10))+P_AXIS), P_AXIS, (((i*((vert-(P_AXIS+125))/10)))+P_AXIS));
    gcvt((each1*i)+limit1,8,number);
    post_textup(pout,font_1st,8,(P_AXIS-11),(((i*((vert-(P_AXIS+125))/10)))+P_AXIS)+(strlen(number)*4/2),number);
    if ((each1*(i+1)) >= i_range) break;
    }
    post_line(pout, (P_AXIS-7), (((i_range/each1)*((vert-(P_AXIS+125))/10))+P_AXIS), P_AXIS, (((i_range/each1)*((vert-(P_AXIS+125))/10))+P_AXIS));
    post_line(pout,P_AXIS,((i*((vert-(P_AXIS+125))/10))+P_AXIS), P_AXIS, (((i_range/each1)*((vert-(P_AXIS+125))/10))+P_AXIS));

if (limits_on)
   {gcvt((float)limit1,8,number);
   post_textup(pout,font_1st,8,(P_AXIS-11),(P_AXIS+(strlen(number)*4)),number);
   }

post_textupcentered(pout,font_1st,10,(P_AXIS-24),(((vert-(P_AXIS+125))/20)*(i_range/each1))+P_AXIS,fasta1);

vert_end=(((i+1)*((vert-(P_AXIS+125))/10))+P_AXIS+100);

if (redo_gif)
    {post_text(pout,font_1st,10,60,(vert_end-90),hh1);
post_text(pout,font_1st,10,70,(vert_end-79),hh2);
post_text(pout,font_1st,10,60,(vert_end-68),hh4);
post_text(pout,font_1st,10,70,(vert_end-57),hh5);
post_text(pout,font_1st,10,60,(vert_end-46),hh7);
sprintf(description,"%.0f%% ",cutoff);
strcat(description,(hh8 + (sizeof(char)*3)));
post_text(pout,font_1st,10,65,(vert_end-35),description);
if (limits_on)
    {sprintf(description,"Only the portion of the graph from %d bp to %d bp of DNA seq. 1\n",limit1,dna1_len);
   post_text(pout,font_1st,10,60,(vert_end-24),description);
    sprintf(description,"and from %d bp to %d bp of DNA seq. 2 are graphed.\n",limit2,dna2_len);
   post_text(pout,font_1st,10,65,(vert_end-13),description);
    }
    }
else
    {
sprintf(description,"The y-axis DNA seq. is %d bp, in file %s:",dna1_len,dna1_file);
post_text(pout,font_1st,10,70,(vert_end-90),description);
post_text(pout,font_1st,10,80,(vert_end-65),fasta1);
sprintf(description,"The x-axis DNA seq. is %d bp, in file %s:",dna2_len,dna2_file);
post_text(pout,font_1st,10,70,(vert_end-50),description);
post_text(pout,font_1st,10,80,(vert_end-35),fasta2);
sprintf(description,"Comparing the DNA seqs. using a %d bp window, returning regions of",window);
post_text(pout,font_1st,10,70,(vert_end-20),description);
sprintf(description,"%.0f%% or greater homology sustained over at least %d bp.\n\n",(cutoff*100),line_min);
post_text(pout,font_1st,10,80,(vert_end-10),description);
    }

time(&bintime);
curtime=localtime(&bintime);
sprintf(description,"%s\n",asctime(curtime));
post_text(pout,font_1st,10,0,0,description);

for (i=1;i<=10;i++)
    {
    post_line(pout,((i*((horz-(P_AXIS*2))/10))+P_AXIS), (P_AXIS-7), ((i*((horz-(P_AXIS*2))/10))+P_AXIS),P_AXIS);
    post_line(pout,(((i-1)*((horz-(P_AXIS*2))/10))+P_AXIS), P_AXIS, ((i*((horz-(P_AXIS*2))/10))+P_AXIS),P_AXIS);
    gcvt((each2*i+limit2),8,number);
    post_text(pout,font_1st,8,((i*((horz-(P_AXIS*2))/10))+P_AXIS)-(strlen(number)*4/2),(P_AXIS-11),number);
    if ((each2*(i+1)) >= j_range) break;
    }
    post_line(pout,(((j_range/each2)*((horz-(P_AXIS*2))/10))+P_AXIS), (P_AXIS-7), (((j_range/each2)*((horz-(P_AXIS*2))/10))+P_AXIS),P_AXIS);
    post_line(pout,((i*((horz-(P_AXIS*2))/10))+P_AXIS), P_AXIS, (((j_range/each2)*((horz-(P_AXIS*2))/10))+P_AXIS),P_AXIS);

if (limits_on)
   {gcvt((float)limit2,8,number);
   post_text(pout,font_1st,8,(P_AXIS-(strlen(number)*4)),(P_AXIS-11),number);
   }

post_textcentered(pout,font_1st,10,(((horz-(P_AXIS*2))/20)*(j_range/each2))+P_AXIS,(P_AXIS-24),fasta2);

for (i=1;i<=100;i++)
    {
    if (((each1/10)*i) >= i_range) break;
    post_line(pout, (P_AXIS-2), ((i*((vert-(P_AXIS+125))/100))+P_AXIS), P_AXIS, ((i*((vert-(P_AXIS+125))/100))+P_AXIS));
    }
for (i=1;i<=100;i++)
    {
    if (((each2/10)*i) >= j_range) break;
    post_line(pout,((i*((horz-(P_AXIS*2))/100))+P_AXIS), (P_AXIS-2), ((i*((horz-(P_AXIS*2))/100))+P_AXIS),P_AXIS);
    }

/*
 *The part that prints out the match lines.
 *
 */

a=0;
for (i1=match[a][0];i1 != 0;i1=match[a][0])
    {b=1;
    for (i2=match[a][b];i2 != 0;i2=match[a][b])
	{h1=post_scaleh(match[a][b+1],each2);h2=post_scaleh(match[a][b+2],each2);
	v1=post_scalev(i1,each1);v2=post_scalev(i2,each1);
	if (squish_on)
	    {
	    if (h1 < 37) {h1=37;};
	    if (v1 < 37) {v1=37;};
	    rect(pout,color[1],h1,(P_AXIS+1),h2,(P_AXIS+1+P_SQUISH_BOX_WIDTH));
	    rect(pout,color[1],(P_AXIS+1),v1,(P_AXIS+1+P_SQUISH_BOX_WIDTH),v2);
	    };
 	post_line(pout,h1,v1,h2,v2);
        b+=3;
        }
    a++;
    }

/*
 *End of page postscript stuff.
 */
fprintf(pout,"showpage\n\n");

        /* Close the file. */
fclose(pout);
}


/*
 *Called by postscript() to print out the commands for drawing a filled 
 *rectange.
 */
void rect(FILE *pout, color3 c3, long int x1, long int y1, long int x2, long int y2)
{
fprintf(pout,"gsave %f %f %f setrgbcolor %d %d %d %d Rec grestore\n",c3.a,c3.b,c3.c,x1-1,y1-1,x2+1,y2+1);
}


/*
 *Called by postscript() to print out the commands for drawing a line.
 */
void post_line(FILE *pout, long int x1, long int y1, long int x2, long int y2)
{
fprintf(pout,"newpath %d %d moveto %d %d lineto stroke\n",x1,y1,x2,y2);
}


/* 
 *Called by postscript() to print out a text string.
 */
void post_text(FILE *pout, char *font_strng, int size, int x, int y, char *text)
{
fprintf(pout,"gsave /%s findfont %d scalefont setfont %d %d moveto [1 0 0 -1 0 0] concat (%s) show grestore\n",font_strng,size,x,y,text);
}

/* 
 *Called by postscript() to print out a text string centered on the passed
 *coord.
 */
void post_textcentered(FILE *pout, char *font_strng, int size, int x, int y, char *text)
{
fprintf(pout,"gsave /%s findfont %d scalefont setfont %d %d moveto [1 0 0 -1 0 0] concat (%s) dup stringwidth pop 2 div neg 0 rmoveto show grestore\n",font_strng,size,x,y,text);
}

/*
 *Called by postscript() to print out a text string vertically.
 */
void post_textup(FILE *pout, char *font_string, int size, int x, int y, char *text)
{
fprintf(pout,"gsave /%s findfont %d scalefont setfont %d %d moveto 90 rotate [-1 0 0 1 0 0] concat (%s) show grestore\n",font_string,size,x,y,text);
}

/*
 *Called by postscript() to print out a text string vertically and centered.
 */
void post_textupcentered(FILE *pout, char *font_string, int size, int x, int y, char *text)
{
fprintf(pout,"gsave /%s findfont %d scalefont setfont %d %d moveto 90 rotate [-1 0 0 1 0 0] concat (%s) dup stringwidth pop 2 div neg 0 rmoveto show grestore\n",font_string,size,x,y,text);
}


float post_scaleh(float x,float each)
{float a;
a=(((((horz-(P_AXIS*2))/10)/each)*(x-limit2))+P_AXIS);
return a;
}

float post_scalev(float x,float each)
{float a;
a=(((((vert-(P_AXIS+125))/10)/each)*(x-limit1))+P_AXIS);
return a;
}

float scaleh(float x,float each)
{float a;
a=(((((horz-90)/10)/each)*(x-limit2))+50);
return a;
}

float scalev(float x,float each)
{float a;
a=(((((vert-175)/10)/each)*(x-limit1))+50);
return a;
}

char *read_dna(char *dna_file, char *dna, long int max_len, char *fasta, long int **feature,char *feat_out, char unknown)
{FILE *fp;
int a=0,c,d,e=0,f,g,h=0;
long int b=0;
char *newp,feat_line[200],in_char;

fp=fopen(dna_file,"r");
if(fp == NULL)
    {
    printf("Can't open file!  Error 4.\n");
    exit (1);
       /* exit or return */
    }
if ((c=getc(fp)) == '>')
    {fasta[a]=c;a++;
    while ((c=getc(fp)) != EOF)
	{if (c == '\n') break;
	fasta[a]=c;a++;
	if (a == 200) 
	    {printf("Error in fasta line in %s.\n",dna_file);
	    exit (1);
	    }
	}
    }
fasta[a]='\0';

/*
 *Parse feature list.  Put pairs of numbers into array feature.
 *This input section isn't rigorous.  Errors or extra spaces in the
 *input could crash it.
 */
while ((c=getc(fp)) == ';')
    {feat_out[h]=c;h++;
    a=0;
    while ((c=getc(fp)) != EOF)
       {feat_out[h]=c;h++;
	if (c == '\n') break;
       feat_line[a]=c;a++;
       if (a == 200) 
    	{printf("Error in feature line in %s.\n",dna_file);
    	exit (1);
    	}
       }
    feat_line[a]='\0';
    a=0;d=0;f=0;g=0;
    while (feat_line[a] != '\0') 
       {if (!(isascii(feat_line[a]) && isdigit(feat_line[a])))
	    {if (f=1)
		{f=0;
		feat_line[a]='\0';
		feature[e][g]=atol(&feat_line[d]);
		d=a+1;
		g++;
		if (g == 3) break;
		}
	    }
	else
	    {if (isascii(feat_line[a]) && isdigit(feat_line[a])) 
		{if (f=0) d=a;
		f=1;
		}
	    }
	a++;
	}
    e++;
    }
 
/*
 *Now input the DNA.
 */
while (c != EOF)
    {if (strrchr(valid,c) != NULL)
	{in_char= tolower(c);
	if ((in_char == 'n') || (in_char == 'x'))
	    in_char=unknown;
	dna[b]= in_char;
	b++;
	if (b == max_len)
            {newp = realloc(dna,(max_len+1000));
        	if(newp != NULL)
                	{dna = newp;
			max_len+=1000;
			}
        	else    {
                	printf("Out of memory!\n");
			exit (1);
                /* exit or return */
                /* but ip still points at 100 ints */
                	}
	    }
	}
    c=getc(fp);
    }
dna[b]='\0';
fclose(fp);
return dna;
}


void initialize(OUT array,int x,int y)
{int a=0,b;
while (a<x)
    {
    b=0;
    while (b<y)
	{array[a][b]=0;
	b++;
	};
    a++;
    };
}
 

float ident_percent(long int i2,long int j2)
{
long int i1=i2-window,j1=j2-window;
int match_now=0, p=0;
float a,b;

while (p<window)
    {if (dna1[i1] == dna2[j1])
        match_now++;
    p++;i1++;j1++;
    }
a=match_now;b=window;
return (a/b);
}

void line(long int i2, long int j2,long int i_end, long int j_end)
{
int b,c,in_a_line=0;
long int start1, start2, temp1, temp2;
float a,matches;

ident=ident_percent(i2,j2);
matches=ident*window;

while ((i2 < (i_end+2)) && (j2 < (j_end+2)))
    {if (in_a_line)
	{if (ident < cutoff)
	    {in_a_line=0;
            if ((i2-start1+window-1) >= line_min)
                {temp1=start1-window+1;
                temp2=start2-window+1;
                if (reloop)
            	    {run_out=1;
                    printf("\t%d\t%d\t%d\t%d\t%d\n",(temp1),(i2-1),(temp2),(j2-1),(i2-temp1));
                    }
                else
                    {
                    b=check_in_match(temp1);
                    c=0;
                    if (b != 0)
                        {b--;
                	while (match[b][c] != 0)
                	    {c++;}
                	c--;
                	if ((c+4) >= match_cmax) more_cols();
                	}
                	else 
                	{b=next_match;next_match++;
                	if (next_match == match_rmax) more_rows();
                	match[b][c]=temp1;
                	}
                    match[b][c+1]=i2-1;
                    match[b][c+2]=temp2;
                    match[b][c+3]=j2-1;
                    }
                }
	    }
	}
    else
	{if (ident >= cutoff)
	    {in_a_line=1;
	    start1=i2;start2=j2;
 	    }
     	}
    a=0;
    if (dna1[i2] == dna2[j2]) a++;
    if (dna1[i2-window] == dna2[j2-window]) a--;
    ident = ((matches+a)/window);
/*
		printf("ident,matches,a: %f, %f, %f\n",ident,matches,a);
*/
    matches += a;i2++;j2++;
    if ((i2 == i_end) || (j2 == j_end)) 
	{ident=0;
	i2++;j2++;
	}
    }
/*
printf("i2,j2,sta,cutoff: %d,%d,%d,%d\n",i2,j2,start1,cutoff);
 */
}

void more_rows()
{int a,b,old_match_rmax=match_rmax;
long int **newp;
newp = realloc(match,((match_rmax+10) * sizeof(long int *)));
                if(newp != NULL)
                        {match = newp;
                        }
                else    {
                        printf("Out of memory in m_r!\n");
                        exit (1);
                /* exit or return */
                /* but ip still points at 100 ints */
                        }
for (a=match_rmax+10;match_rmax<a;match_rmax++)
    {match[match_rmax] = malloc(match_cmax * sizeof(long int));
    if(match[match_rmax] == NULL)
        {
        printf("\nFailure to allocate memory for growing match!\n");
        exit(1);
	}
    }
for (a=old_match_rmax;a<match_rmax;a++)
    {for (b=0;b < match_cmax;b++)
	{match[a][b]=0;
	}
    }
}

void more_cols()
{int a,b;
long int *newp;
for (a=0;a<match_rmax;a++)
    {
    newp = realloc(match[a],((match_cmax+9) * sizeof(long int)));
                if(newp != NULL)
                        {match[a] = newp;
			for (b=match_cmax;b < (match_cmax+9);b++)
			    {match[a][b]=0;
			    }
                        }
                else    {
                        printf("Out of memory in m_c!, col: %d\n",a);
                        exit (1);
                /* exit or return */
                /* but ip still points at 100 ints */
                        }
    }
match_cmax+=9;
}


int check_in_match(long int line_start)
{int a=0;
while (match[a][0] != 0)
    {if (match[a][0] == line_start) return (a+1);
    a++;
    }
return 0;
}

void print_match(void)
{int a=0,b;
long int i1,i2;

printf("The first  DNA seq. is %d bp, in file %s:\n%s\n",dna1_len,dna1_file,fasta1);
printf("%s\n",feat_out1);
printf("The second DNA seq. is %d bp, in file %s:\n%s\n",dna2_len,dna2_file,fasta2);
printf("%s\n",feat_out2);
printf("Comparing the DNA seqs. using a %d bp window, returning regions of\n",window);
printf("%.0f%% or greater homology sustained over at least %d bp.\n\n",(cutoff*100),line_min);
printf("\tDNA seq. 1       DNA seq. 2\n");
printf("      Line     Line    Line     Line   Match\n");
printf("      Start    End     Start    End    Length\n");
printf("---------------------------------------------\n");
 
for (i1=match[a][0];i1 != 0;i1=match[a][0])
    {b=1;
    for (i2=match[a][b];i2 != 0;i2=match[a][b])
	{printf("\t%d\t%d\t%d\t%d\t%d\n",i1,i2,match[a][b+1],match[a][b+2],(i2-i1+1));
	b+=3;
	}
    a++;
    };

if (match[0][0] == 0) printf("No matches.\n");
}
/*
end of script
*/

